from django.conf import settings
from django.db import models
from django.utils.translation import gettext, gettext_lazy as _
import sys

from .signals import price_pre_save_signal


CURRENCIES = [(currency_title.lower(), currency_title) 
               for currency_title in settings.MULTICURRENCY_CURRENCIES]


def create_multicurrency_price_model(shared_model, related_name, **kwargs):
    meta = {}
    capitalized_field_name = ''.join([ 
                            s.capitalize() for s in related_name.split('_') ])

    if shared_model._meta.abstract:
        raise TypeError("Can't create MulticurrencyPriceModel for " + 
                        "abstract class {0}".format(shared_model.__name__))

    meta['app_label'] = shared_model._meta.app_label
    meta['db_tablespace'] = shared_model._meta.db_tablespace
    meta['managed'] = shared_model._meta.managed
    meta.setdefault('db_table', f'{shared_model._meta.db_table}_{related_name}')
    meta.setdefault('default_permissions', ())

    name = str(f'{shared_model.__name__}{capitalized_field_name}')

    attrs = {}
    additional_params = ['verbose_name','blank','null']
    attrs['Meta'] = type(str('Meta'), (object,), meta)
    attrs['__module__'] = shared_model.__module__
    attrs['objects'] = models.Manager()
    _parent_relations_params={
        'to':shared_model, 
        'editable':False, 
        'on_delete':models.CASCADE,
        'related_name':related_name
    }
    for key, value in kwargs.items():
        if key not in additional_params:
            raise AttributeError(f"Unknown attribute {key}")
        _parent_relations_params[key] = value
    attrs['master'] = models.OneToOneField(**_parent_relations_params)

    # add columns for each currency
    for currency in settings.MULTICURRENCY_CURRENCIES:
        attrs[f'{currency.lower()}_amount'] = models.FloatField(
                                                    _('Amount'), 
                                                    blank=True, 
                                                    null=False)

    price_model = models.base.ModelBase(name, (MulticurrencyPriceModel,), attrs)

    mod = sys.modules[shared_model.__module__]
    setattr(mod, name, price_model)
    models.signals.pre_save.connect(price_pre_save_signal, price_model)

    return price_model


class MulticurrencyPriceModel(models.Model, metaclass=models.base.ModelBase):
    currency = models.CharField(_('Currency'),
                                max_length=3,
                                choices=CURRENCIES,
                                db_index=True)

    def __str__(self) -> str:
        title = f"{self.pk}"
        for currency in CURRENCIES:
            currency_price = getattr(self,f'{currency[0]}_amount')
            title += f" {currency[1]}({currency_price})"
        return title

    class Meta:
        abstract = True
        default_permissions = ()


class MulticurrencyPriceModelMixin(models.Model):
    def _fetch_price_field_model(self, field:str) -> MulticurrencyPriceModel:
        price_field = getattr(self._meta.model, field)
        if not isinstance(price_field, 
                models.fields.related_descriptors.ReverseOneToOneDescriptor):
            raise AttributeError(f"Price field must be a instance of " + 
                "models.fields.related_descriptors.ReverseOneToOneDescriptor")

        fields_list = self._meta.get_fields(include_hidden=True)
        for _field in fields_list:
            if _field.name == field: return _field.field.model
        raise AttributeError(f"No price field with title {field}")

    def _fetch_price_instance(self, field:str) -> MulticurrencyPriceModel:
        price_model = self._fetch_price_field_model(field)
        return price_model.objects.get(master=self)

    def create_price(self, field:str, currency:str, amount:float) -> \
                                                    MulticurrencyPriceModel:
        '''
            Create price related to parent
            Params:
                field - price field title
                currency - capitalized main currency from settings currency list
                amount - amount of price in main currency
            Result:
                Instance of recently created MulticurrencyPriceModel
        '''
        currency_field = f'{currency.lower()}_amount'
        price_model = self._fetch_price_field_model(field)

        return price_model.objects.create(**{
            'master':self,
            'currency':currency,
            currency_field:amount
        })

    def update_price(self, field:str, currency:str, amount:float) -> \
                                                    MulticurrencyPriceModel:
        '''
            Update price related to parent
            Params:
                field - price field title
                currency - capitalized main currency from settings currency list
                amount - amount of price in main currency
            Result:
                Instance of recently edited MulticurrencyPriceModel
        '''
        currency_field = f'{currency.lower()}_amount'
        price_instance = self._fetch_price_instance(field)
        price_instance.currency = currency
        setattr(price_instance,currency_field,amount)
        price_instance.save()

        return price_instance

    def delete_price(self, field:str) -> None:
        '''
            Delete price related to parent
            Params:
                field - price field title
            Result:
                None if success
        '''
        price_instance = self._fetch_price_instance(field)
        price_instance.delete()

    class Meta:
        abstract = True
        default_permissions = ()


from django.conf import settings

from .exceptions import MulticurrencyException
from .utils import import_class


def price_pre_save_signal(sender, instance, raw, **kwargs):
    Exchanger = import_class(settings.MULTICURRENCY_EXCHANGER)
    if not hasattr(Exchanger, 'get_exchange_rate'):
        raise MulticurrencyException(f'Please provide "get_exchange_rate" method for exchanger {Exchanger}')

    main_currency = instance.currency
    amount = getattr(instance, f'{main_currency.lower()}_amount', None)

    if not amount:
        raise MulticurrencyException(f'Please, provide amount in {main_currency}')

    exchanger = Exchanger()

    currencies_list = list(settings.MULTICURRENCY_CURRENCIES)
    currencies_list.remove(main_currency.upper())

    for currency in currencies_list:

        exchange_rate = exchanger.get_exchange_rate(main_currency.lower(), 
                                                        currency.lower())
        if not exchange_rate:   
            exchange_rate = exchanger.get_exchange_rate(currency.lower(), 
                                                        main_currency.lower()) 
            if not exchange_rate:
                raise MulticurrencyException(
            f'No exchange rate for pair {main_currency.upper()}/{currency} or {currency}/{main_currency.upper()}')
            
            exchange_rate = 1/exchange_rate

        setattr(instance, f'{currency.lower()}_amount', amount*exchange_rate)


from django.db import models

from .models import create_multicurrency_price_model, MulticurrencyPriceModel


class MultiCurrencyPriceField:
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def contribute_to_class(self, cls, name, **kwargs):
        self.price_model = \
                create_multicurrency_price_model(cls, name, **self._kwargs)


import importlib

def import_class(class_path:str) -> object:
    """
        Function return class by string path:
            class_path: module1.sub_module.file.MyClass
        return MyClass
    """
    # get split module path and class name from class path
    module_path, class_name = \
        class_path[:class_path.rfind('.')], class_path[class_path.rfind('.')+1:]

    # import module and return class from module
    module = importlib.import_module(module_path)
    return getattr(module, class_name)
    


class MulticurrencyException(Exception):
    def __init__(self, message=None):
        self.message = message if message else ''

    def __str__(self) -> str:
        return f'Multicurrency Exception {self.message}'


#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path
import codecs


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='django-multicurrency',
    version='0.1.2',
    license='Apache 2.0',

    requires=[
        'Django (>=3.0)',
    ],

    description='Simple Django multicurrency field',
    long_description=long_description,
    long_description_content_type="text/markdown",

    author='Neural Dynamics',
    author_email='neuraldynamics.web@gmail.com',

    url='https://github.com/NeuralDynamicsWeb/django-multicurrency',
    download_url = 'https://github.com/NeuralDynamicsWeb/django-multicurrency/archive/refs/tags/v_02.1.tar.gz',

    packages=find_packages(exclude=('example*',)),
    include_package_data=True,

    keywords = ['Django', 'Multicurrency'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6',
)